from requests import Session
import requests
from geovisio_cli.model import Geovisio
from geovisio_cli.exception import raise_for_status
from geovisio_cli import utils
import os
from dataclasses import dataclass, field
from typing import List, Optional
import toml
from rich import print


def login(s: Session, geovisio: Geovisio) -> bool:
    """
    Login to geovisio and store auth cookie in session
    """
    if geovisio.token is not None:
        return _login_with_token(s, geovisio)

    return _login_with_stored_credentials(s, geovisio)


def _login_with_token(s: Session, geovisio: Geovisio) -> bool:
    s.headers.update({"Authorization": f"Bearer {geovisio.token}"})
    return True


def _login_with_stored_credentials(s: Session, geovisio: Geovisio):
    creds = _read_existing_credentials(geovisio)

    instance_cred = creds.get_instance_credentials(geovisio.url)
    if not instance_cred:
        _generate_and_update_credentials(creds, geovisio)
        return False

    account_name = _check_if_associated(geovisio, instance_cred)
    if not account_name:
        claim_url = f"{geovisio.url}/api/auth/tokens/{instance_cred.token_id}/claim"
        print(
            f"🔐 We're waiting for you to link your user account with generated API token. To finalize authentication, please either go to the URL below, or scan the QR code below."
        )
        print(f"[green]{claim_url}[/green]")
        _display_qr_code(claim_url)
        return False

    print(f"👤 Using stored credentials, logged in as [green]{account_name}[/green]")
    s.headers.update({"Authorization": f"Bearer {instance_cred.jwt_token}"})
    return True


@dataclass
class InstanceCredential:
    url: str
    jwt_token: str
    token_id: str

    def get_claim_url(self) -> str:
        return f"{self.url}/api/auth/tokens/{self.token_id}/claim"


@dataclass
class Credentials:
    instances: List[InstanceCredential] = field(default_factory=lambda: [])

    def from_toml(self, data):
        self.instances = [
            InstanceCredential(
                url=i["url"], jwt_token=i["jwt_token"], token_id=i["token_id"]
            )
            for i in data.get("instances", [])
        ]

    def toml(self):
        return {
            "instances": [
                {"url": i.url, "jwt_token": i.jwt_token, "token_id": i.token_id}
                for i in self.instances
            ]
        }

    def get_instance_credentials(self, instance_url: str):
        return next((c for c in self.instances if c.url == instance_url), None)


def create_auth_credentials(geovisio: Geovisio) -> InstanceCredential:
    """
    Login to geovisio and store auth cookie in session
    """
    utils.test_geovisio_url(geovisio.url)
    creds = _read_existing_credentials(geovisio)

    instance_cred = creds.get_instance_credentials(geovisio.url)
    if instance_cred:
        account_name = _check_if_associated(geovisio, instance_cred)
        if account_name:
            print(f"👤 Already logged to instance as [green]{account_name}[/green]")
            return instance_cred
        claim_url = instance_cred.get_claim_url()
        print(
            f"🔐 We're waiting for you to link your user account with generated API token. To finalize authentication, please either go to the URL below, or scan the QR code below."
        )
        print(f"[green]{claim_url}[/green]")
        _display_qr_code(claim_url)
        return instance_cred

    return _generate_and_update_credentials(creds, geovisio)


def _generate_and_update_credentials(
    creds: Credentials, geovisio: Geovisio
) -> InstanceCredential:
    i = _generate_new_instance_credentials(geovisio)

    creds.instances.append(i)

    _write_credentials(creds)
    return i


def get_config_file_path() -> str:
    # store config file either if [XDG](https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html) defined directory or in a user specifig .config directory
    config_file_dir = os.environ.get("XDG_CONFIG_HOME") or os.path.join(
        os.path.expanduser("~"), ".config"
    )

    return os.path.join(config_file_dir, "geovisio", "config.toml")


def _read_existing_credentials(geovisio: Geovisio) -> Credentials:
    creds = Credentials()
    if not os.path.exists(get_config_file_path()):
        return creds

    with open(get_config_file_path(), "r") as f:
        tomlFileContent = f.read()
        creds.from_toml(toml.loads(tomlFileContent))
        f.close()
    return creds


def _generate_new_instance_credentials(geovisio: Geovisio) -> InstanceCredential:
    token_response = requests.post(f"{geovisio.url}/api/auth/tokens/generate")
    raise_for_status(token_response, "Impossible to generate a GeoVisio token")

    token = token_response.json()
    jwt_token = token["jwt_token"]
    id = token["id"]

    claim_url = next(l["href"] for l in token["links"] if l["rel"] == "claim")
    print(
        f"🔐 Your computer is not yet authorized against GeoVisio API {geovisio.url}. To authenticate, please either go to the URL below, or scan the QR code below."
    )
    print(f"[green]{claim_url}[/green]")
    _display_qr_code(claim_url)
    return InstanceCredential(url=geovisio.url, jwt_token=jwt_token, token_id=id)


def _write_credentials(creds: Credentials):
    config_file_path = get_config_file_path()
    os.makedirs(os.path.dirname(config_file_path), exist_ok=True)
    with open(config_file_path, "w") as f:
        f.write(toml.dumps(creds.toml()))
        f.close()


def _check_if_associated(
    geovisio: Geovisio, creds: InstanceCredential
) -> Optional[str]:
    token_response = requests.get(
        f"{geovisio.url}/api/users/me",
        headers={"Authorization": f"Bearer {creds.jwt_token}"},
    )
    if token_response.status_code == 403 or token_response.status_code == 401:
        return None
    raise_for_status(token_response, "Impossible to get token status")

    return token_response.json()["name"]


def _display_qr_code(url):
    import qrcode  # type: ignore
    import io

    qr = qrcode.QRCode()
    qr.add_data(url)
    f = io.StringIO()
    qr.print_ascii(out=f)
    f.seek(0)
    print(f.read())



import pytest
import os
from geovisio_cli import sequence, exception, model
from .conftest import FIXTURE_DIR, MOCK_API_URL
from pathlib import Path
from geopic_tag_reader import reader
import toml
import requests
from requests_mock import Adapter


@pytest.mark.parametrize(
    ("data", "method", "expected"),
    (
        (["1.jpg", "2.jpg", "3.jpg"], "filename-asc", ["1.jpg", "2.jpg", "3.jpg"]),
        (["3.jpg", "1.jpg", "2.jpg"], "filename-asc", ["1.jpg", "2.jpg", "3.jpg"]),
        (["3.jpg", "1.jpg", "2.jpeg"], "filename-asc", ["1.jpg", "2.jpeg", "3.jpg"]),
        (["10.jpg", "5.jpg", "1.jpg"], "filename-asc", ["1.jpg", "5.jpg", "10.jpg"]),
        (["C.jpg", "A.jpg", "B.jpg"], "filename-asc", ["A.jpg", "B.jpg", "C.jpg"]),
        (
            ["CAM1_001.jpg", "CAM2_002.jpg", "CAM1_002.jpg"],
            "filename-asc",
            ["CAM1_001.jpg", "CAM1_002.jpg", "CAM2_002.jpg"],
        ),
        (["1.jpg", "2.jpg", "3.jpg"], "filename-desc", ["3.jpg", "2.jpg", "1.jpg"]),
        (["3.jpg", "1.jpg", "2.jpg"], "filename-desc", ["3.jpg", "2.jpg", "1.jpg"]),
        (["3.jpg", "1.jpg", "2.jpeg"], "filename-desc", ["3.jpg", "2.jpeg", "1.jpg"]),
        (["10.jpg", "5.jpg", "1.jpg"], "filename-desc", ["10.jpg", "5.jpg", "1.jpg"]),
        (["C.jpg", "A.jpg", "B.jpg"], "filename-desc", ["C.jpg", "B.jpg", "A.jpg"]),
        (
            ["CAM1_001.jpg", "CAM2_002.jpg", "CAM1_002.jpg"],
            "filename-desc",
            ["CAM2_002.jpg", "CAM1_002.jpg", "CAM1_001.jpg"],
        ),
    ),
)
def test_sort_files_names(data, method, expected):
    dataPictures = [sequence.Picture(path=p) for p in data]
    resPictures = sequence._sort_files(dataPictures, method)
    assert expected == [pic.path for pic in resPictures]


@pytest.mark.parametrize(
    ("data", "method", "expected"),
    (
        (
            [["1.jpg", 1], ["2.jpg", 2], ["3.jpg", 3]],
            "time-asc",
            ["1.jpg", "2.jpg", "3.jpg"],
        ),
        (
            [["1.jpg", 2], ["2.jpg", 3], ["3.jpg", 1]],
            "time-asc",
            ["3.jpg", "1.jpg", "2.jpg"],
        ),
        (
            [["1.jpg", 1.01], ["2.jpg", 1.02], ["3.jpg", 1.03]],
            "time-asc",
            ["1.jpg", "2.jpg", "3.jpg"],
        ),
        (
            [["1.jpg", 1], ["2.jpg", 2], ["3.jpg", 3]],
            "time-desc",
            ["3.jpg", "2.jpg", "1.jpg"],
        ),
        (
            [["1.jpg", 2], ["2.jpg", 3], ["3.jpg", 1]],
            "time-desc",
            ["2.jpg", "1.jpg", "3.jpg"],
        ),
        (
            [["1.jpg", 1.01], ["2.jpg", 1.02], ["3.jpg", 1.03]],
            "time-desc",
            ["3.jpg", "2.jpg", "1.jpg"],
        ),
    ),
)
def test_sort_files_time(data, method, expected):
    dataPictures = []
    for p in data:
        name, ts = p
        m = reader.GeoPicTags(47.7, -1.78, ts, 0, "flat", "Panoramax", "180++", 4)
        dataPictures.append(sequence.Picture(path=name, metadata=m))

    resPictures = sequence._sort_files(dataPictures, method)
    assert expected == [pic.path for pic in resPictures]


def test_rw_sequence_toml(tmp_path):
    s = sequence.Sequence(
        title="SEQUENCE",
        id="blab-blabla-blablabla",
        path=str(tmp_path),
        pictures=[
            sequence.Picture(
                id="blou-bloublou-bloubloublou-1", path=str(tmp_path / "1.jpg")
            ),
            sequence.Picture(
                id="blou-bloublou-bloubloublou-2", path=str(tmp_path / "2.jpg")
            ),
            sequence.Picture(
                id="blou-bloublou-bloubloublou-3", path=str(tmp_path / "3.jpg")
            ),
        ],
        sort_method=sequence.SortMethod.time_desc,
    )
    res = sequence._write_sequence_toml(s)
    assert res == str(tmp_path / "_geovisio.toml")
    res2 = sequence._update_sequence_from_toml(sequence.Sequence(path=str(tmp_path)))
    assert s == res2
    assert res2.sort_method == sequence.SortMethod.time_desc


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
)
def test_read_sequence(datafiles):
    # First read : position is based on picture names
    seq = sequence._read_sequence(Path(datafiles))
    seqTomlPath = sequence._write_sequence_toml(seq)

    assert os.path.isfile(seqTomlPath)

    # Edit TOML file : position is inverted
    with open(seqTomlPath, "r+") as seqToml:
        editedSeqToml = seqToml.read()
        editedSeqToml = (
            editedSeqToml.replace("position = 1", "position = A")
            .replace("position = 2", "position = 1")
            .replace("position = A", "position = 2")
        )
        seqToml.seek(0)
        seqToml.write(editedSeqToml)
        seqToml.close()

        # Read sequence 2 : position should match edited TOML
        seq = sequence._read_sequence(Path(datafiles))
        assert seq.pictures[0].path.endswith("e2.jpg")
        assert seq.pictures[1].path.endswith("e1.jpg")


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
    os.path.join(FIXTURE_DIR, "invalid_pic.jpg"),
)
def test_read_sequence_invalid_file(datafiles):
    # Read sequence from files
    seq = sequence._read_sequence(Path(datafiles))
    seqTomlPath = sequence._write_sequence_toml(seq)

    assert os.path.isfile(seqTomlPath)

    # Check if invalid_pic is marked as broken
    with open(seqTomlPath, "r") as seqToml:
        seq2toml = seqToml.read()
        seq2 = sequence.Sequence()
        seq2.from_toml(toml.loads(seq2toml))
        assert len(seq2.pictures) == 4
        assert seq2.pictures[0].status == "broken-metadata"
        assert seq2.pictures[1].status is None
        assert seq2.pictures[2].status is None
        assert seq2.pictures[3].status is None


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
)
def test_read_sequence_sort_method_changed_set2unset(datafiles):
    # Write toml with sort method defined
    seq = sequence._read_sequence(
        Path(datafiles), sortMethod=sequence.SortMethod.time_desc
    )
    seqTomlPath = sequence._write_sequence_toml(seq)
    assert os.path.isfile(seqTomlPath)

    # Read sequence from toml without sort method = should reuse read one
    seq = sequence._read_sequence(Path(datafiles))
    assert seq.pictures[0].path.endswith("e2.jpg")
    assert seq.pictures[1].path.endswith("e1.jpg")


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
)
def test_read_sequence_sort_method_changed_different(datafiles):
    # Write toml with sort method defined
    seq = sequence._read_sequence(
        Path(datafiles), sortMethod=sequence.SortMethod.time_desc
    )
    seqTomlPath = sequence._write_sequence_toml(seq)
    assert os.path.isfile(seqTomlPath)

    # Read sequence from toml without sort method = should reuse read one
    with pytest.raises(exception.CliException) as e:
        seq = sequence._read_sequence(
            Path(datafiles), sortMethod=sequence.SortMethod.filename_asc
        )

    assert e.match("Sort method passed as argument")


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "invalid_pic.jpg"),
)
def test_upload_with_no_valid_file(datafiles):
    with pytest.raises(exception.CliException) as e:
        seq = sequence._read_sequence(Path(datafiles))

    assert e.match("All read pictures have invalid metadata")


def mock_api_post_collection_fail(requests_mock):
    requests_mock.post(
        MOCK_API_URL + "/api/collections",
        exc=requests.exceptions.ConnectTimeout,
    )


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
)
def test_upload_collection_create_failure(requests_mock, datafiles):
    mock_api_post_collection_fail(requests_mock)

    with pytest.raises(exception.CliException) as e:
        sequence.upload(
            path=datafiles,
            geovisio=model.Geovisio(url=MOCK_API_URL),
            title="Test",
            alreadyBlurred=True,
        )

    assert str(e.value).startswith("Error while connecting to the API")


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
)
def test_upload_with_invalid_file(requests_mock, datafiles):
    # Put apart third picture
    os.rename(datafiles + "/e2.jpg", datafiles + "/e2.bak")
    os.rename(datafiles + "/e3.jpg", datafiles + "/e3.bak")

    # Mock collection creation
    gvsMock = model.Geovisio(url=MOCK_API_URL)
    seqId = "123456789"
    picId1 = "123"
    picId2 = "456"
    picId3 = "789"
    requests_mock.get(f"{MOCK_API_URL}/api", json={})
    requests_mock.get(f"{MOCK_API_URL}/api/configuration", json={})
    requests_mock.get(
        f"{MOCK_API_URL}/api/collections/{seqId}",
        json={"id": seqId, "title": "whatever"},
    )
    requests_mock.post(
        f"{MOCK_API_URL}/api/collections",
        json={"id": seqId},
        headers={"Location": f"{MOCK_API_URL}/api/collections/{seqId}"},
    )
    requests_mock.post(
        f"{MOCK_API_URL}/api/collections/{seqId}/items",
        json={"type": "Feature", "id": picId1},
        headers={"Location": f"{MOCK_API_URL}/api/collections/{seqId}/items/{picId1}"},
    )
    uploadReport = sequence.upload(path=Path(datafiles), geovisio=gvsMock, title=None)

    # Check previous pictures are OK
    assert len(uploadReport.uploaded_pictures) == 1
    assert len(uploadReport.errors) == 0

    # Make other pictures available
    os.rename(datafiles + "/e2.bak", datafiles + "/e2.jpg")
    os.rename(datafiles + "/e3.bak", datafiles + "/e3.jpg")
    with open(datafiles + "/_geovisio.toml") as f:
        seq = toml.loads(f.read())
        seq["pictures"]["e2.jpg"] = {"path": str(datafiles) + "/e2.jpg", "position": 2}
        seq["pictures"]["e3.jpg"] = {"path": str(datafiles) + "/e3.jpg", "position": 3}
        f.close()

    with open(datafiles + "/_geovisio.toml", "w") as f2:
        f2.write(toml.dumps(seq))
        f2.close()

        # Mock item call to fail
        requests_mock.post(
            f"{MOCK_API_URL}/api/collections/{seqId}/items",
            [
                {
                    "json": {"type": "Feature", "id": picId2},
                    "status_code": 202,
                    "headers": {
                        "Location": f"{MOCK_API_URL}/api/collections/{seqId}/items/{picId2}"
                    },
                },
                {"status_code": 500},
            ],
        )
        uploadReport2 = sequence.upload(
            path=Path(datafiles), geovisio=gvsMock, title=None
        )

        assert len(uploadReport2.uploaded_pictures) == 1
        assert len(uploadReport2.errors) == 1



import os

FIXTURE_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "fixtures")
MOCK_API_URL = "http://geovisio-local.dev"


def pytest_addoption(parser):
    parser.addoption("--external-geovisio-url", action="store")



import pytest
import os
from geovisio_cli import sequence
from .conftest import FIXTURE_DIR
from pathlib import Path, PurePath


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
    os.path.join(FIXTURE_DIR, "not_a_pic.md"),
)
def test_upload_with_invalid_file(datafiles):
    s = sequence.process(path=Path(datafiles), title=None)

    assert len(s.pictures) == 3
    assert [PurePath(p.path).stem for p in s.pictures] == ["e1", "e2", "e3"]
    assert s.title == Path(datafiles).name



import typer
from typing_extensions import Annotated
from pathlib import Path
from geovisio_cli import sequence, exception, model, auth, utils, __version__
from rich import print
from rich.panel import Panel
from typing import Optional
import os


app = typer.Typer()


def version_callback(value: bool):
    if value:
        print(f"GeoVisio command-line client ([blue bold]v{__version__}[/blue bold])")
        utils.check_if_lastest_version()
        raise typer.Exit()


@app.callback(help=f"GeoVisio command-line client (v{__version__})")
def common(
    version: Annotated[
        Optional[bool],
        typer.Option(
            "--version",
            callback=version_callback,
            is_eager=True,
            help="Show GeoVisio command-line client version and exit",
        ),
    ] = None,
):
    pass


@app.command()
def upload(
    path: Path = typer.Argument(..., help="Local path to your sequence folder"),
    api_url: str = typer.Option(..., help="GeoVisio endpoint URL"),
    user: str = typer.Option(
        default=None,
        hidden=True,
        help="""DEPRECATED: GeoVisio user name if the geovisio instance needs it.
If none is provided and the geovisio instance requires it, the username will be asked during run.
""",
        envvar="GEOVISIO_USER",
    ),
    password: str = typer.Option(
        default=None,
        hidden=True,
        help="""DEPRECATED: GeoVisio password if the geovisio instance needs it.
If none is provided and the geovisio instance requires it, the password will be asked during run.
Note: is is advised to wait for prompt without using this variable.
""",
        envvar="GEOVISIO_PASSWORD",
    ),
    wait: bool = typer.Option(default=False, help="Wait for all pictures to be ready"),
    isBlurred: bool = typer.Option(
        False,
        "--is-blurred/--is-not-blurred",
        help="Define if sequence is already blurred or not",
    ),
    title: Optional[str] = typer.Option(
        default=None,
        help="Collection title. If not provided, the title will be the directory name.",
    ),
    token: Optional[str] = typer.Option(
        default=None,
        help="""GeoVisio token if the geovisio instance needs it.

If none is provided and the geovisio instance requires it, the token will be asked during run.
Note: is is advised to wait for prompt without using this variable.
""",
    ),
    sort_method: Optional[sequence.SortMethod] = typer.Option(
        default="time-asc",
        help="Strategy used for sorting your pictures. Either by filename or EXIF time, in ascending or descending order.",
    ),
):
    """Processes and sends a given sequence on your GeoVisio API"""

    def cmd():
        if user or password:
            raise exception.CliException(
                "user/password authentication have been deprecated, use a token or `geovisio login` instead"
            )
        geovisio = model.Geovisio(url=api_url, token=token)
        sequence.upload(
            path,
            geovisio,
            wait=wait,
            alreadyBlurred=isBlurred,
            title=title,
            sortMethod=sort_method,
        )

    _run_command(cmd, "importing collection")


@app.command()
def test_process(
    path: Path = typer.Argument(..., help="Local path to your sequence folder"),
    title: Optional[str] = typer.Option(
        default=None,
        help="Collection title. If not provided, the title will be the directory name.",
    ),
    sort_method: Optional[sequence.SortMethod] = typer.Option(
        default="time-asc",
        help="Strategy used for sorting your pictures. Either by filename or EXIF time, in ascending or descending order.",
    ),
):
    """(For testing) Generates a TOML file with metadata used for upload"""

    def cmd():
        sequence.process(path, title, sortMethod=sort_method)
        outputFile = os.path.join(path, sequence.SEQUENCE_TOML_FILE)
        print(
            "✅ [green]Metadata file saved to: [bold]" + outputFile + "[/bold][/green]"
        )

    _run_command(cmd, "processing collection")


@app.command()
def collection_status(
    id: Optional[str] = typer.Option(default=None, help="Id of the collection"),
    api_url: Optional[str] = typer.Option(default=None, help="GeoVisio endpoint URL"),
    location: Optional[str] = typer.Option(
        default=None, help="Full url of the collection"
    ),
    wait: bool = typer.Option(default=False, help="wait for all pictures to be ready"),
):
    """
    Print the status of a collection.\n
    Either a --location should be provided, with the full location url of the collection
    or only the --id combined with the --api-url
    """

    def cmd():
        if location is None:
            if api_url is None or id is None:
                raise exception.CliException(
                    "The way to identify the collection should be either with --location or with --id combined with --api-url"
                )
            l = f"{api_url}/api/collections/{id}"
        else:
            l = location

        mySequence = sequence.Sequence(id=id, location=l)
        sequence.display_sequence_status(mySequence)

        if wait:
            sequence.wait_for_sequence(mySequence)

    _run_command(cmd, "getting collection status")


@app.command(
    help=f"""
    Authenticate into the given instance, and save credentials in a configuration file.

    This will generate credentials, and ask the user to visit a page to associate those credentials to the user's account.

    The credentials will be stored in {auth.get_config_file_path()}
    """
)
def login(
    api_url: str = typer.Option(..., help="GeoVisio endpoint URL"),
):
    return _run_command(
        lambda: auth.create_auth_credentials(model.Geovisio(url=api_url)),
        "authenticating",
    )


def _run_command(command, command_name_for_error):
    try:
        utils.check_if_lastest_version()
        command()
    except exception.CliException as e:
        print(
            Panel(
                f"{e}",
                title=f"[red]Error while {command_name_for_error}",
                border_style="red",
            )
        )
        return 1



from geovisio_cli import utils
import geovisio_cli
import pytest


@pytest.mark.parametrize(
    ("local_version", "pypi_version", "up_to_date"),
    (
        ("0.1.0", "0.2.0", False),
        ("0.2.0", "0.2.1", False),
        ("0.2.0", "0.2.0", True),
        ("0.3.0", "0.2.0", True),
        ("1.0.0", "0.2.0", True),
    ),
)
def test_check_if_lastest_version(
    requests_mock, local_version, pypi_version, up_to_date
):
    sub_pypi_response = {"info": {"version": pypi_version}}
    geovisio_cli.__version__ = local_version
    requests_mock.get("https://pypi.org/pypi/geovisio_cli/json", json=sub_pypi_response)
    assert utils.check_if_lastest_version() == up_to_date



from geovisio_cli.exception import CliException
from rich import print
import requests

REQUESTS_TIMEOUT = 30


def test_geovisio_url(geovisio: str):
    full_url = f"{geovisio}/api"
    try:
        r = requests.get(full_url, timeout=REQUESTS_TIMEOUT)
    except (
        requests.Timeout,
        requests.ConnectionError,
        requests.ConnectTimeout,
        requests.TooManyRedirects,
    ) as e:
        raise CliException(
            f"""The API is not reachable. Please check error and used URL below, and retry later if the URL is correct.

[bold]Used URL:[/bold] {full_url}
[bold]Error:[/bold]
{e}"""
        )
    except Exception as e:
        raise CliException(
            f"""Error while connecting to the API. Please check error and used URL below

[bold]Used URL:[/bold] {full_url}
[bold]Error:[/bold]
{e}"""
        )

    if r.status_code == 404:
        raise CliException(
            f"""The API URL is not valid.

Note that your URL should be the API root (something like https://geovisio.fr, https://panoramax.ign.fr or any other geovisio instance).
Please make sure you gave the correct URL and retry.

[bold]Used URL:[/bold] {full_url}
[bold]Error:[/bold]
{r.text}"""
        )
    if r.status_code > 404:
        raise CliException(
            f"""The API is unavailable for now. Please check given error and retry later.
[bold]Used URL:[/bold] {full_url}
[bold]Error[/bold] (code [cyan]{r.status_code}[/cyan]):
{r.text}"""
        )


def check_if_lastest_version():
    from packaging import version
    import geovisio_cli

    pypi_url = "https://pypi.org/pypi/geovisio_cli"
    response = requests.get(f"{pypi_url}/json")
    latest_version = response.json()["info"]["version"]

    if version.parse(latest_version) > version.parse(geovisio_cli.__version__):
        print(
            f"⚠️ A newer geovisio_cli version {latest_version} is available on PyPI (available on {pypi_url}).\nWe highly recommend updating as this tool is still in beta version, and new versions ensure good compatibility with GeoVisio API."
        )
        return False
    return True



from requests import Response, HTTPError


class CliException(Exception):
    def __init__(self, msg):
        super().__init__(msg)


def raise_for_status(r: Response, msg: str):
    try:
        r.raise_for_status()
    except HTTPError as e:
        raise CliException(msg) from e



from geovisio_cli import auth, sequence, exception, model
import os
import pytest
from pathlib import Path
from ..conftest import FIXTURE_DIR
from .conftest import _login
import requests


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
)
def test_auth_with_token(geovisio_with_token, datafiles):
    collection = sequence.upload(
        path=Path(datafiles), geovisio=geovisio_with_token, title="some title"
    )

    assert len(collection.uploaded_pictures) == 3
    assert len(collection.errors) == 0


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
)
def test_login(geovisio, datafiles, user_credential, tmp_path):
    assert not os.path.exists(tmp_path / "geovisio" / "config.toml")
    l = auth.create_auth_credentials(geovisio)

    # we call the auth credentials while loggin
    with requests.session() as s:
        _login(s, geovisio, user_credential)
        claim = s.get(l.get_claim_url())
        assert claim.status_code == 200
        assert (
            claim.text == "You are now logged in the CLI, you can upload your pictures"
        )

    assert os.path.exists(tmp_path / "geovisio" / "config.toml")

    # doing a geovisio upload should work without crendentials now
    uploadReport = sequence.upload(
        path=Path(datafiles), geovisio=geovisio, title="some title"
    )
    assert len(uploadReport.uploaded_pictures) == 3
    assert len(uploadReport.errors) == 0


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
)
def test_login_without_claim(geovisio, datafiles, tmp_path):
    """Login without claiming the token should result to errors for pictures upload"""
    assert not os.path.exists(tmp_path / "geovisio" / "config.toml")
    l = auth.create_auth_credentials(geovisio)

    # a config file should have been created
    assert os.path.exists(tmp_path / "geovisio" / "config.toml")

    # doing a geovisio upload should not work as the token is not usable yet
    with pytest.raises(
        exception.CliException,
        match="🔁 Computer not authenticated yet, impossible to upload pictures, but you can try again the same upload command after finalizing the login",
    ):
        sequence.upload(path=Path(datafiles), geovisio=geovisio, title="some title")


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
)
def test_login_on_invalid_url_path(geovisio_with_token, datafiles):
    with pytest.raises(exception.CliException) as e:
        auth.create_auth_credentials(
            model.Geovisio(url=geovisio_with_token.url + "/some_additional_path")
        )
    msg = str(e.value)
    assert msg.startswith(
        f"""The API URL is not valid.

Note that your URL should be the API root (something like https://geovisio.fr, https://panoramax.ign.fr or any other geovisio instance).
Please make sure you gave the correct URL and retry.

[bold]Used URL:[/bold] {geovisio_with_token.url}/some_additional_path/api
[bold]Error:[/bold]"""
    )



from pathlib import Path, PurePath
from dataclasses import dataclass, field
from typing import List, Optional, Union
import requests
from rich import print
from rich.table import Table
from rich.markup import escape
from rich.progress import (
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeElapsedColumn,
    BarColumn,
    MofNCompleteColumn,
)
from rich.panel import Panel
from rich.console import Group
from rich.live import Live
from geovisio_cli.exception import CliException, raise_for_status
from geovisio_cli.auth import login
from geovisio_cli.model import Geovisio
from geovisio_cli import utils
from geopic_tag_reader import reader
from PIL import Image
from time import sleep
from datetime import timedelta
import os
import toml
from . import exception
from enum import Enum
import logging

SEQUENCE_TOML_FILE = "_geovisio.toml"


class SortMethod(str, Enum):
    filename_asc = "filename-asc"
    filename_desc = "filename-desc"
    time_asc = "time-asc"
    time_desc = "time-desc"


@dataclass
class InteriorOrientation:
    make: str
    model: str
    field_of_view: Optional[int]


@dataclass
class Picture:
    path: Optional[str] = None
    id: Optional[str] = None
    location: Optional[str] = None
    status: Optional[str] = None
    metadata: Optional[reader.GeoPicTags] = None

    def toml(self):
        return {
            "path": self.path,
            "id": self.id,
            "location": self.location,
            "status": self.status,
        }

    @staticmethod
    def from_toml(data):
        return Picture(
            path=data.get("path"),
            id=data.get("id"),
            location=data.get("location"),
            status=data.get("status"),
        )


@dataclass
class Sequence:
    title: str = ""
    path: str = ""
    id: Optional[str] = None
    location: Optional[str] = None
    producer: Optional[str] = None
    interior_orientation: List[InteriorOrientation] = field(default_factory=lambda: [])
    pictures: List[Picture] = field(default_factory=lambda: [])
    sort_method: Optional[SortMethod] = None

    def toml(self):
        res = {
            "sequence": {
                "title": self.title,
                "path": self.path,
                "id": self.id,
                "location": self.location,
                "producer": self.producer,
                "sort_method": self.sort_method.value
                if self.sort_method is not None
                else None,
            },
            "pictures": {},
        }

        for pos, pic in enumerate(self.pictures, start=1):
            pict = pic.toml()
            pict["position"] = pos
            res["pictures"][os.path.basename(pic.path)] = pict

        return res

    def from_toml(self, data):
        if data.get("sequence"):
            self.title = data["sequence"].get("title", "")
            self.path = data["sequence"].get("path", "")
            self.id = data["sequence"].get("id")
            self.location = data["sequence"].get("location")
            self.producer = data["sequence"].get("producer")
            self.sort_method = (
                SortMethod(data["sequence"]["sort_method"])
                if "sort_method" in data["sequence"]
                else None
            )

        if data.get("pictures"):
            self.pictures = [
                Picture.from_toml(picData)
                for picId, picData in sorted(
                    data["pictures"].items(), key=lambda item: int(item[1]["position"])
                )
            ]

    def all_done(self):
        return self.nb_waiting() + self.nb_preparing() == 0

    def nb_ready(self):
        return sum((1 for p in self.pictures if p.status == "ready"))

    def nb_waiting(self):
        return sum((1 for p in self.pictures if p.status == "waiting-for-process"))

    def nb_preparing(self):
        return sum((1 for p in self.pictures if p.status.startswith("preparing")))

    def nb_broken(self):
        return sum((1 for p in self.pictures if p.status == "broken"))


@dataclass
class UploadError:
    position: int
    picture_path: str
    error: Union[str, dict]
    status_code: int


@dataclass
class UploadReport:
    location: str
    uploaded_pictures: List[Picture] = field(default_factory=lambda: [])
    errors: List[UploadError] = field(default_factory=lambda: [])


def process(
    path: Path, title: Optional[str], sortMethod: Optional[SortMethod] = None
) -> Sequence:
    sequence = _read_sequence(path, title, sortMethod)
    _check_sequence(sequence)
    _write_sequence_toml(sequence)
    return sequence


def upload(
    path: Path,
    geovisio: Geovisio,
    title: Optional[str],
    wait: bool = False,
    alreadyBlurred: bool = False,
    sortMethod: Optional[SortMethod] = None,
) -> UploadReport:
    # early test that the given url is correct
    utils.test_geovisio_url(geovisio.url)
    with requests.session() as s:
        # early test login
        if not _login_if_needed(s, geovisio):
            raise CliException(
                "🔁 Computer not authenticated yet, impossible to upload pictures, but you can try again the same upload command after finalizing the login"
            )

        sequence = process(path, title, sortMethod)

        return _publish(s, sequence, geovisio, wait, alreadyBlurred)


def _publish(
    session: requests.Session,
    sequence: Sequence,
    geovisio: Geovisio,
    wait: bool,
    alreadyBlurred: bool,
) -> UploadReport:
    # Read sequence data
    if sequence.id:
        sequence = info(sequence)
        print(
            f'🔄 Syncing collection "{sequence.title}" (local folder: {sequence.path}, API ID: {sequence.id})'
        )
    else:
        print(f'📂 Publishing collection "{sequence.title}" ({sequence.path})')

    data = {}
    if sequence.title:
        data["title"] = sequence.title

    # List pictures to upload
    picturesToUpload = {}
    for i, p in enumerate(sequence.pictures, start=1):
        if p.id is None or p.status == "broken":
            picturesToUpload[i] = p

    # Create sequence on initial publishing
    if not sequence.id:
        seq = session.post(
            f"{geovisio.url}/api/collections", data=data, timeout=utils.REQUESTS_TIMEOUT
        )
        raise_for_status(seq, "Impossible to query GeoVisio")

        sequence.id = seq.json()["id"]
        sequence.location = seq.headers["Location"]
        _write_sequence_toml(sequence)

        print(f"✅ Created collection {sequence.location}")

    else:
        if len(picturesToUpload) == 0:
            print(
                f"✅ Everything ({len(sequence.pictures)} picture{'s' if len(sequence.pictures) != 1 else ''}) have already been uploaded, nothing to do"
            )
            assert sequence.location
            return UploadReport(location=sequence.location)
        print(
            f"⏭️ Skipping {len(sequence.pictures) - len(picturesToUpload)} already published pictures"
        )

    if not sequence.location:
        raise CliException("Sequence has no API location defined")

    report = UploadReport(location=sequence.location)

    uploading_progress = Progress(
        TextColumn("{task.description}"),
        BarColumn(),
        TimeElapsedColumn(),
        TextColumn("[{task.completed}/{task.total}]"),
    )
    current_pic_progress = Progress(
        TextColumn("📷 Processing [bold purple]{task.fields[file]}"),
        SpinnerColumn("simpleDots"),
    )
    error_progress = Progress(TextColumn("{task.description}"))

    last_error = Progress(
        TextColumn("🔎 Last error 🔎\n{task.description}"),
    )
    error_panel = Panel(Group(error_progress, last_error), title="Errors")
    uploading_task = uploading_progress.add_task(
        f"[green]🚀 Uploading pictures...",
        total=len(picturesToUpload),
    )
    current_pic_task = current_pic_progress.add_task("", file="")
    progress_group = Group(
        uploading_progress,
        current_pic_progress,
        error_panel,
    )
    error_task = error_progress.add_task("[green]No errors")
    last_error_task = last_error.add_task("", visible=False)
    with Live(progress_group):
        for i, p in picturesToUpload.items():
            if not p.path:
                raise CliException(f"Missing path for picture {i}")

            uploading_progress.advance(uploading_task)
            current_pic_progress.update(current_pic_task, file=p.path.split("/")[-1])
            try:
                picture_response = session.post(
                    f"{sequence.location}/items",
                    files={"picture": open(p.path, "rb")},
                    data={
                        "position": i,
                        "isBlurred": "true" if alreadyBlurred else "false",
                    },
                    timeout=utils.REQUESTS_TIMEOUT,
                )
            except (requests.Timeout,) as timeout_error:
                raise CliException(
                    f"""Timeout while trying to post picture. Maybe the instance is overloaded? Please contact your instance administrator.

            [bold]Error:[/bold]
            {timeout_error}"""
                )
            except (
                requests.ConnectionError,
                requests.ConnectTimeout,
                requests.TooManyRedirects,
            ) as cnx_error:
                raise CliException(
                    f"""Impossible to reach GeoVisio while trying to post a picture, connection was lost. Please contact your instance administrator.

            [bold]Error:[/bold]
            {cnx_error}"""
                )

            # Picture at given position exists -> mark it as OK
            if picture_response.status_code == 409:
                sequence = status(sequence)
                _write_sequence_toml(sequence)
                report.uploaded_pictures.append(p)

            elif picture_response.status_code >= 400:
                body = (
                    picture_response.json()
                    if picture_response.headers.get("Content-Type")
                    == "application/json"
                    else picture_response.text
                )
                report.errors.append(
                    UploadError(
                        position=i,
                        picture_path=p.path,
                        status_code=picture_response.status_code,
                        error=body,
                    )
                )

                error_progress.update(
                    error_task,
                    description=f"[bold red]{len(report.errors)} errors",
                )
                last_error.update(last_error_task, description=body, visible=True)
                p.status = "broken"
                _write_sequence_toml(sequence)

            else:
                p.location = picture_response.headers["Location"]
                p.id = picture_response.json()["id"]
                p.status = None
                report.uploaded_pictures.append(p)
                _write_sequence_toml(sequence)

    if not report.uploaded_pictures:
        print(
            f"[red]💥 All pictures upload of sequence {sequence.title} failed! 💥[/red]"
        )
    else:
        print(
            f"🎉 [bold green]{len(report.uploaded_pictures)}[/bold green] pictures uploaded"
        )
    if report.errors:
        print(f"[bold red]{len(report.errors)}[/bold red] pictures not uploaded:")
        for e in report.errors:
            msg: Union[str, dict] = e.error
            if isinstance(e.error, str):
                msg = escape(e.error.replace("\n", "\\n"))
            print(f" - {e.picture_path} (status [bold]{e.status_code}[/]): {msg}")

    if wait:
        wait_for_sequence(sequence)
    else:
        print(f"Note: You can follow the picture processing with the command:")
        from rich.syntax import Syntax

        print(f"[bold]geovisio collection-status --wait --location {sequence.location}")
    return report


def _check_sequence(sequence: Sequence):
    if not sequence.pictures:
        raise CliException(f"No pictures to upload for sequence {sequence.title}")


def _sort_files(
    pictures: List[Picture], method: Optional[SortMethod] = SortMethod.time_asc
) -> List[Picture]:
    """Sorts pictures according to their file name

    Parameters
    ----------
    pictures : Picture[]
        List of pictures to sort
    method : SortMethod
        Sort logic to adopt (time-asc, time-desc, filename-asc, filename-desc)

    Returns
    -------
    Picture[]
        List of pictures, sorted
    """

    if method is None:
        method = SortMethod.time_asc

    if method not in [item.value for item in SortMethod]:
        raise exception.CliException("Invalid sort strategy: " + str(method))

    # Get the sort logic
    strat, order = method.split("-")

    # Sort based on filename
    if strat == "filename":

        def sort_fct(p):
            try:  # Try to sort based on numeric notation
                return int(PurePath(p.path or "").stem)
            except:  # Otherwise, sort as strings
                return PurePath(p.path or "").stem

        pictures.sort(key=sort_fct)

    # Sort based on picture ts
    elif strat == "time":
        pictures.sort(key=lambda p: p.metadata.ts if p.metadata is not None else 0)

    if order == "desc":
        pictures.reverse()

    return pictures


def _read_sequence(
    path: Path, title: Optional[str] = None, sortMethod: Optional[SortMethod] = None
) -> Sequence:
    if not path.is_dir():
        raise CliException(f"{path} is not a directory, cannot read pictures")

    if title is None:
        title = path.name

    s = Sequence(title=title, path=str(path.absolute()), sort_method=sortMethod)

    # Check if a TOML file exists, then use it instead of generating one
    stoml = os.path.join(s.path, SEQUENCE_TOML_FILE)
    if os.path.isfile(stoml):
        print(f"📄 Using metadata from existing config file: {stoml}")
        _update_sequence_from_toml(s)

        # Check sort method
        if (
            sortMethod is not None
            and s.sort_method is not None
            and sortMethod != s.sort_method
        ):
            raise CliException(
                f'Sort method passed as argument ({sortMethod.value}) is different from the one defined in your metadata file ({s.sort_method.value}).\nYou may either change --sort-method argument from command-line, or change "sort_method" in your metadata file.'
            )

    # Create sequence from pictures files
    else:
        print(f"🔍 Listing available pictures")
        for f in path.iterdir():
            if not f.is_file():
                continue
            if f.suffix.lower() not in [".jpg", ".jpeg"]:
                continue

            # Read metadata
            try:
                meta = reader.readPictureMetadata(Image.open(str(f)))
                s.pictures.append(Picture(path=str(f), metadata=meta))
            except Exception as e:
                logging.warning(f"Picture {str(f)} has invalid metadata: {str(e)}")
                s.pictures.append(Picture(path=str(f), status="broken-metadata"))

        s.pictures = _sort_files(s.pictures, sortMethod)

    # Check if at least one picture is valid
    if (
        len(s.pictures) > 0
        and len([p for p in s.pictures if p.status != "broken-metadata"]) == 0
    ):
        raise CliException(
            "❌ All read pictures have invalid metadata.\nPlease check if your pictures are geolocated and have a date defined.\nFor more information: https://gitlab.com/geovisio/api/-/blob/develop/docs/15_Pictures_requirements.md"
        )

    return s


def _write_sequence_toml(sequence: Sequence):
    """Writes TOML sequence metadata file"""

    outputFile = os.path.join(sequence.path, SEQUENCE_TOML_FILE)
    with open(outputFile, "w") as f:
        f.write(toml.dumps(sequence.toml()))
        f.close()

    return outputFile


def _update_sequence_from_toml(sequence: Sequence) -> Sequence:
    """Updates given sequence based on its TOML config file"""

    tomlFile = os.path.join(sequence.path, SEQUENCE_TOML_FILE)
    with open(tomlFile, "r") as f:
        tomlFileContent = f.read()
        sequence.from_toml(toml.loads(tomlFileContent))
        f.close()

    return sequence


def _login_if_needed(session: requests.Session, geovisio: Geovisio) -> bool:
    # Check if API needs login
    apiConf = session.get(f"{geovisio.url}/api/configuration")
    if apiConf.json().get("auth", {}).get("enabled", False):
        logged_in = login(session, geovisio)
        if not logged_in:
            return False
    return True


def status(sequence: Sequence) -> Sequence:
    s = requests.get(
        f"{sequence.location}/geovisio_status", timeout=utils.REQUESTS_TIMEOUT
    )
    if s.status_code == 404:
        raise CliException(f"Sequence {sequence.location} not found")
    if s.status_code >= 400:
        raise CliException(
            f"Impossible to get sequence {sequence.location} status: {s.text}"
        )
    r = s.json()

    if len(sequence.pictures) == 0:
        sequence.pictures = [
            Picture(id=p["id"], status=p["status"]) for p in r["items"]
        ]
    else:
        for i, p in enumerate(r["items"]):
            sequence.pictures[i].id = p["id"]
            sequence.pictures[i].status = p["status"]

    return sequence


def info(sequence: Sequence) -> Sequence:
    if not sequence.location:
        raise CliException(f"Sequence has no location set")

    s = requests.get(sequence.location, timeout=utils.REQUESTS_TIMEOUT)
    if s.status_code == 404:
        raise CliException(f"Sequence {sequence.location} not found")
    if s.status_code >= 400:
        raise CliException(
            f"Impossible to get sequence {sequence.location} status: {s.text}"
        )
    r = s.json()
    producer = next(
        (p["name"] for p in r.get("providers", []) if "producer" in p["roles"]), None
    )
    summary = r.get("summaries", {}).get("pers:interior_orientation", [])

    sequence.id = r["id"]
    sequence.title = r["title"]
    sequence.producer = producer
    sequence.interior_orientation = [
        InteriorOrientation(
            make=s.get("make"),
            model=s.get("model"),
            field_of_view=s.get("field_of_view"),
        )
        for s in summary
    ]

    return sequence


def display_sequence_status(sequence: Sequence):
    seq_status = status(sequence)
    seq_info = info(sequence)

    s = f"Sequence [bold]{seq_info.title}[/bold]"
    if seq_info.producer is not None:
        s += f" produced by [bold]{seq_info.producer}[/bold]"
    s += " taken with"
    for i in seq_info.interior_orientation:
        s += f" [bold]{i.make} {i.model}[/bold]"
        if i.field_of_view:
            s += f" ({i.field_of_view}°)"

    print(s)
    table = Table()

    table.add_column("Total")
    table.add_column("Ready", style="green")
    table.add_column("Waiting", style="magenta")
    table.add_column("Preparing", style="magenta")
    table.add_column("Broken", style="red")

    table.add_row(
        f"{len(seq_status.pictures)}",
        f"{seq_status.nb_ready()}",
        f"{seq_status.nb_waiting()}",
        f"{seq_status.nb_preparing()}",
        f"{seq_status.nb_broken()}",
    )
    print(table)


def _print_final_sequence_status(sequence: Sequence):
    nb_broken = sequence.nb_broken()
    nb_ready = sequence.nb_ready()
    if nb_ready == 0:
        print(f"[red]💥 No picture processed")
        return
    s = f"✅ {nb_ready} pictures processed"
    if nb_broken:
        s += f"([red]{nb_broken}[/red] cannot be processed)"
    print(s)


def wait_for_sequence(sequence: Sequence, timeout: Optional[timedelta] = None):
    seq_status = status(sequence)
    if seq_status.all_done():
        _print_final_sequence_status(seq_status)
        return

    print("🔭 Waiting for pictures to be processed by geovisio")
    status_progress = Progress(
        TextColumn("{task.description}"),
        BarColumn(),
        TimeElapsedColumn(),
        MofNCompleteColumn(),
        "•",
        TextColumn("{task.fields[processing]}"),
    )
    processing_task = status_progress.add_task(
        f"[green]⏳ Processing ...",
        total=1,
        processing="",
    )
    progress_group = Group(
        status_progress,
    )
    waiting_time = timedelta(seconds=2)
    elapsed = timedelta(0)

    with Live(progress_group):
        while True:
            # TODO: display some stats about those errors

            nb_preparing = seq_status.nb_preparing()
            nb_waiting = seq_status.nb_waiting()
            processing = f"{nb_preparing} picture{('s' if nb_preparing != 0 else '')} currently processed"
            status_progress.update(
                processing_task,
                total=len(seq_status.pictures),
                completed=seq_status.nb_ready(),
                processing=processing,
            )

            if nb_waiting + nb_preparing == 0:
                break

            elapsed += waiting_time
            if timeout is not None and elapsed > timeout:
                raise CliException(f"❌ Sequence not ready after {elapsed}, stoping")

            sleep(waiting_time.total_seconds())
            seq_status = status(sequence)

    _print_final_sequence_status(seq_status)



import os
import pytest
from ..conftest import FIXTURE_DIR
from pathlib import Path
import requests
from geovisio_cli import sequence, exception, model
from datetime import timedelta
import toml


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
)
def test_valid_upload(geovisio_with_token, datafiles):
    uploadReport = sequence.upload(
        path=Path(datafiles), geovisio=geovisio_with_token, title="some title"
    )
    assert len(uploadReport.uploaded_pictures) == 3
    assert len(uploadReport.errors) == 0

    collection = sequence.Sequence(location=uploadReport.location)
    sequence.wait_for_sequence(collection, timeout=timedelta(minutes=1))
    status = sequence.status(collection)
    sequence_info = sequence.info(collection)
    # 3 pictures should have been uploaded
    assert len(status.pictures) == 3

    for i in status.pictures:
        assert i.status == "ready"

    # the collection should also have 3 items
    collection = requests.get(f"{collection.location}/items")
    collection.raise_for_status()

    features = collection.json()["features"]
    assert len(features) == 3

    assert sequence_info.title == "some title"
    assert sequence_info.interior_orientation == [
        sequence.InteriorOrientation(
            make="SONY", model="FDR-X1000V", field_of_view=None
        )
    ]

    # Check generated TOML file
    tomlFile = datafiles / sequence.SEQUENCE_TOML_FILE
    with open(tomlFile, "r") as f:
        tomlFileContent = f.read()
        seqToml = toml.loads(tomlFileContent)
        assert seqToml["sequence"]["id"] == sequence_info.id
        assert seqToml["pictures"]["e1.jpg"]["id"] == features[0]["id"]
        assert seqToml["pictures"]["e2.jpg"]["id"] == features[1]["id"]
        assert seqToml["pictures"]["e3.jpg"]["id"] == features[2]["id"]


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
)
def test_resume_upload(geovisio_with_token, datafiles):
    # Make e2 not valid to have a partial upload
    picE2 = datafiles / "e2.jpg"
    picE2bak = datafiles / "e2.bak"
    os.rename(picE2, picE2bak)
    with open(picE2, "w") as picE2file:
        picE2file.write("")
        picE2file.close()

    # Start upload -> 2 uploaded pics + 1 failure
    uploadReport = sequence.upload(
        path=Path(datafiles),
        geovisio=geovisio_with_token,
        title="some title",
        sortMethod=sequence.SortMethod.filename_asc,
    )
    assert len(uploadReport.uploaded_pictures) == 2
    assert len(uploadReport.errors) == 1
    assert uploadReport.errors[0].position == 2

    # Check TOML file -> e2 has no ID but broken status
    tomlFile = datafiles / sequence.SEQUENCE_TOML_FILE
    with open(tomlFile, "r") as f:
        tomlFileContent = f.read()
        f.close()
        seqToml = toml.loads(tomlFileContent)
        assert seqToml["pictures"]["e2.jpg"].get("id") is None
        assert seqToml["pictures"]["e2.jpg"]["status"] == "broken"

    # Make e2 valid
    os.remove(picE2)
    os.rename(picE2bak, picE2)

    # Launch again upload : 1 uploaded pic + 0 failure
    uploadReport2 = sequence.upload(
        path=Path(datafiles),
        geovisio=geovisio_with_token,
        title="some title",
        sortMethod=sequence.SortMethod.filename_asc,
    )
    assert uploadReport2.location == uploadReport.location
    assert len(uploadReport2.uploaded_pictures) == 1
    assert len(uploadReport2.errors) == 0

    # Check TOML file -> everything has ID and looks like a charm
    with open(tomlFile, "r") as f:
        tomlFileContent = f.read()
        f.close()
        seqToml2 = toml.loads(tomlFileContent)
        assert seqToml2["sequence"]["id"] == seqToml["sequence"]["id"]
        assert (
            seqToml2["pictures"]["e1.jpg"]["id"] == seqToml["pictures"]["e1.jpg"]["id"]
        )
        assert seqToml2["pictures"]["e2.jpg"]["id"] is not None
        assert seqToml2["pictures"]["e2.jpg"].get("status") is None
        assert (
            seqToml2["pictures"]["e3.jpg"]["id"] == seqToml["pictures"]["e3.jpg"]["id"]
        )


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
)
def test_upload_twice(geovisio_with_token, datafiles):
    """Uploading twice the same sequence, should result of nothing done in the second upload"""
    # Start upload -> 2 uploaded pics + 1 failure
    uploadReport = sequence.upload(
        path=Path(datafiles), geovisio=geovisio_with_token, title="some title"
    )
    assert len(uploadReport.uploaded_pictures) == 3
    assert len(uploadReport.errors) == 0

    # Launch again upload : 1 uploaded pic + 0 failure
    uploadReport2 = sequence.upload(
        path=Path(datafiles), geovisio=geovisio_with_token, title="some title"
    )
    assert uploadReport2.location == uploadReport.location
    assert len(uploadReport2.uploaded_pictures) == 0
    assert len(uploadReport2.errors) == 0


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
)
def test_upload_on_invalid_url_host(datafiles):
    with pytest.raises(exception.CliException) as e:
        sequence.upload(
            path=Path(datafiles),
            geovisio=model.Geovisio(url="http://some_invalid_url"),
            title="some title",
        )
    msg = str(e.value)
    assert msg.startswith(
        """The API is not reachable. Please check error and used URL below, and retry later if the URL is correct.

[bold]Used URL:[/bold] http://some_invalid_url/api
[bold]Error:[/bold]"""
    )
    assert "Failed to establish a new connection:" in msg


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
)
def test_upload_on_invalid_url_path(geovisio_with_token, datafiles):
    with pytest.raises(exception.CliException) as e:
        sequence.upload(
            path=Path(datafiles),
            geovisio=model.Geovisio(
                url=geovisio_with_token.url + "/some_additional_path"
            ),
            title=None,
        )
    msg = str(e.value)
    assert msg.startswith(
        f"""The API URL is not valid.

Note that your URL should be the API root (something like https://geovisio.fr, https://panoramax.ign.fr or any other geovisio instance).
Please make sure you gave the correct URL and retry.

[bold]Used URL:[/bold] {geovisio_with_token.url}/some_additional_path/api
[bold]Error:[/bold]"""
    )


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
)
def test_upload_on_invalid_url_schema(datafiles):
    with pytest.raises(exception.CliException) as e:
        sequence.upload(
            path=Path(datafiles),
            geovisio=model.Geovisio(url="a non valid url at all"),
            title=None,
        )
    assert str(e.value).startswith(
        """Error while connecting to the API. Please check error and used URL below

[bold]Used URL:[/bold] a non valid url at all/api
[bold]Error:[/bold]"""
    )



import pytest
import os
from geovisio_cli.model import Geovisio
import requests


@pytest.fixture(scope="session")
def geovisio(pytestconfig):
    """
    If --external-geovisio-url has been given to pytest use an already running geovisio, else spawn a fully configured geovisio for integration tests
    """
    external_geovisio_url = pytestconfig.getoption("--external-geovisio-url")
    if external_geovisio_url:
        yield Geovisio(url=external_geovisio_url)
        return

    from testcontainers import compose

    dco_file = os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        "docker-compose-geovisio.yml",
    )
    with compose.DockerCompose(
        ".",
        compose_file_name=dco_file,
        pull=True,
    ) as compose:
        port = compose.get_service_port("geovisio-api", 5000)
        api_url = f"http://api.geovisio.localtest.me:{port}"
        compose.wait_for(api_url)

        yield Geovisio(url=api_url)
        stdout, stderr = compose.get_logs()
        if stderr:
            print("Errors\n:{}".format(stderr))


@pytest.fixture(scope="session")
def user_credential():
    """Credential of a fake created account on keycloak"""
    return ("elysee", "my password")


@pytest.fixture(scope="session")
def geovisio_with_token(geovisio, user_credential):
    token = _get_token(geovisio, user_credential)
    return Geovisio(
        url=geovisio.url,
        token=token,
    )


def _get_token(geovisio, user_credential):
    with requests.session() as s:
        _login(s, geovisio, user_credential)
        tokens = s.get(f"{geovisio.url}/api/users/me/tokens")
        tokens.raise_for_status()
        token_link = next(
            t["href"] for t in tokens.json()[0]["links"] if t["rel"] == "self"
        )
        assert token_link
        jwt_token = s.get(token_link)
        jwt_token.raise_for_status()
        return jwt_token.json()["jwt_token"]


def _login(session, geovisio, user_credential):
    login = session.get(f"{geovisio.url}/api/auth/login")

    url = _get_keycloak_authenticate_form_url(login)

    r = session.post(
        url,
        data={"username": user_credential[0], "password": user_credential[1]},
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        allow_redirects=True,
    )

    # a bit hacky, but since for the moment we only submit a form to keycloak, to know if the login was successful,
    # we need to check that we were redirected to geovisio
    r.raise_for_status()
    assert r.history != 0


def _get_keycloak_authenticate_form_url(response):
    """Little hack to parse keycloak HTML to get the url to the authenticate form"""
    import re

    url = re.search('action="(.*login-actions/authenticate[^"]*)"', response.text)

    assert url, f"impossible to find form action in keycloak response: {response.text}"
    url = url.group(1).replace("&amp;", "&")
    return url


@pytest.fixture(scope="function", autouse=True)
def override_config_home(tmp_path):
    """Set XDG_CONFIG_HOME to temporary directory, so tests newer write a real user config file"""
    old_var = os.environ.get("XDG_CONFIG_HOME")

    os.environ["XDG_CONFIG_HOME"] = str(tmp_path)
    yield

    if old_var:
        os.environ["XDG_CONFIG_HOME"] = old_var
    else:
        del os.environ["XDG_CONFIG_HOME"]



import pytest
from geovisio_cli import sequence, exception


def test_status_on_unknown_collection(geovisio):
    with pytest.raises(exception.CliException) as e:
        sequence.status(
            sequence.Sequence(location=f"{geovisio.url}/api/collections/some_bad_id")
        )
    assert e.match(f"Sequence {geovisio.url}/api/collections/some_bad_id not found")
